// Aspell Data Objects Implementations 
// Copyright 1998 Under the terms of the LGPL

//
// Aspell's main word list data is stored in 4 large blocks of memory
//
// * The Word Hash Table
// * The Word List
// * The Soundslike Hash Table
// * The Soundslike List
//
// 1a) The Word Hash Table
// This consists of an open address hash table which contains pointers
// to the actual words in the word list
//
// 1b) The Word List
// This consists of the actual word list and is layed out as follows:
//   <Word1><null char><Word2><null char>...
//
// 2a) The Soundslike Hash Table
// This consists of an open address hash table which contains pointers
// to a soundslike object.
//
// 2b) The Soundslike Object
// The soundslike object is layed out as follow:
//  <Word1 pointer><Word2 pointer>...<Num of Words><Soundslike><null char>
//  <const char *><const char *>...<unsigned short int><char[]><char>
// The pointer to the object points to the beginning of the Soundslike string
// The Word pointers consists of the the words which have the same 
//   soundslike pattern
//
// 2c) The Soundslike List
// This consists of Soundslike Objects back to back:
//  <Soundslike object 1><Soundslike object 2> ...
// There is no delimiter between the objects
//
//
//                          Format of the *.wrd files
//
// (This part is in ascii format)
// <"master_wl"><ws><lang name><ws><# words><ws>
//     <hash size><ws><size of list block><\n>
// (The rest is in binary format>
// <Wordlist>
// <Word Hash Table>
//
// The word hash table is a vector of unsigned its which contains an offset
// of where they can be found in the word list.
//
//                          Format of the *.sl files
//
// (This part is in ascii format)
// <"master_wl"><ws><lang name><ws><# words><ws>
//     <hash size><ws><size of list block><\n>
// (The rest is in binary format>
// <Soundslike object list>
// <Soundslike Hash Table>
//
// Soundslike oject is laid out as follows:
//   <Num of Words><Word 1 offset>...<Soundslike><\0>
//   <unsigned short int><unsigned int>...<char[]><char>
// And like the .wrd file the hash table contains offsets not pointers.
//

#include <string>
#include <fstream.h>
#include <strstream.h>

#include <time.h>
#include <sys/stat.h>

#include "aspell/i_error.hh"
#include "aspell/data.hh"
#include "data.hh"
#include "language.hh"
#include "hash-t.hh"

#define try_file(name) \
  { path = (name); \
    FILE.open( path.c_str() ); \
    if ( !(FILE.rdstate() & ios::badbit)) \
      return path;}

string find_file(const string &file, const string &suffix_, 
		 const string &try_dir_, bool suffix_must = false) {
  ifstream FILE;
  
  string suffix;
  if (!suffix_must && 
      file.substr(file.size()-suffix_.size(),suffix_.size()) == suffix_)
    suffix = "";
  else 
    suffix = suffix_;
  string try_dir;
  if (file[0] == '/')
    try_dir = "";
  else
    try_dir = try_dir_ + '/';

  string path;
  
  if (!suffix_must) 		 
    try_file(file);
  try_file(file + suffix);
  if (!suffix_must)
    try_file(try_dir + file);
  try_file(try_dir + file + suffix);

  return path;
} 

string find_data_file(const string &file, const string &suffix) {
  const char *temp = getenv("ASPELL_DATA_DIR");
  if (temp)
    return find_file(file, suffix, temp, true);
  else
    return find_file(file, suffix, LIBDIR, true);
}

string find_per_file(const string &file, const string &suffix) {
 const char *temp = getenv("HOME");
  if (temp) 
    return find_file(file, suffix, temp);
  else 
    return find_file(file, suffix, "");
}

void SC_ReadOnly_WordList::load(const string &file_name) {
  
  ifstream WORDS((file_name).c_str(), ios::in | ios::bin);
  if ( WORDS.rdstate() & ios::badbit )
    throw ae_cant_read(file_name, 0);

  string check_word;
  WORDS >> check_word;
  if (check_word != "master_wl")
    throw ae_bad_format(file_name, 0);

  WORDS >> check_word;

  string lang_dir;
  string::size_type spos = file_name.rfind('/');
  if (spos == string::npos) {
    lang_dir = "";
  } else {
    lang_dir = file_name.substr(0,spos+1);
  }

  try {
    parent_->set_check_lang(check_word,lang_dir);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, 0);
  }

  int count, bucket, size;
  WORDS >> count >> bucket >> size;
  WORDS.ignore();
  master_ = new char[size];

  WORDS.read(master_, size);

  word_list.table().resize(bucket);
  WordList::table_type::iterator i   = word_list.table().begin();

  unsigned int pos;
  while (!WORDS.eof()) {
    WORDS.read(&pos, sizeof(unsigned int));
    *i = master_ + pos;
    ++i;
  }

  word_list.set_size(count);
}

void SC_ReadOnly_Soundslikes::load(const string &file_name, const char *d)
{
  ifstream SOUNDS((file_name).c_str(), ios::in, ios::bin);
  if ( SOUNDS.rdstate() & ios::badbit )
    throw ae_cant_read(file_name, 0);
  
  string check_word;
  SOUNDS >> check_word;
  if (check_word != "master_ml") 
    throw ae_bad_format(file_name, 0);

  SOUNDS >> check_word;

  string lang_dir;
  string::size_type spos = file_name.rfind('/');
  if (spos == string::npos) {
    lang_dir = "";
  } else {
    lang_dir = file_name.substr(0,spos+1);
  }
  
  try {
    parent_->set_check_lang(check_word,lang_dir);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, 0);
  }

  int count, bucket, size;
  SOUNDS >> count >> bucket >> size;
  main_data = new char[size];
  empty_wordlist = main_data + sizeof(unsigned short int);
  char *data = main_data;
  
  SOUNDS.ignore();

  *reinterpret_cast<unsigned short int *>(data) = 0;
  *(data + sizeof(unsigned short int *)) = '\0';
  data = main_data + sizeof(const char *);
  
  unsigned short int list_count;
  unsigned int       pos;
  unsigned int       m;
  while (SOUNDS.read(reinterpret_cast<void*>(&list_count),sizeof(unsigned short int)),
	 list_count) {
    for(unsigned short int i=0; i != list_count; ++i) {
      SOUNDS.read(&pos, sizeof(unsigned int));
      *reinterpret_cast<const char **>(data) = d + pos;
      data += sizeof(const char *);
    }
    *reinterpret_cast<unsigned short int *>(data) = list_count;
    data += sizeof(unsigned short int);
    while ((*data = SOUNDS.get()) != '\0') ++data;
    ++data;
    m = (data - main_data) % sizeof(const char *);
    if (m != 0)
      data += sizeof(const char *) - m;
  }

  lookup_table.table().resize(bucket);

  LookupTable::table_type::iterator i   = lookup_table.table().begin();

  while (!SOUNDS.eof()) {
    SOUNDS.read(&pos, sizeof(unsigned int));
    *i = main_data + pos;
    ++i;
  }

  lookup_table.set_size(count);
}

const SC_ReadOnly_Soundslikes::WordList SC_ReadOnly_Soundslikes
::words_w_soundslike(const char *word) const {
  LookupTable::const_iterator i = lookup_table.find(word);
  if (i == lookup_table.end() || !(*i).size()) { 
    //                                  ^^ FIXME
    // There should not be a need for the size check as there should never be
    // any empty lists in the lookup_table.
    return empty_wordlist;
  } else {                                       
    return *i;
  }
}

const SC_ReadOnly_Soundslikes::WordList SC_ReadOnly_Soundslikes
::words_w_soundslike(const char *word, SC_Sound &M) const {
  LookupTable::const_iterator i = lookup_table.find(word);
  if (i == lookup_table.end() || !(*i).size()) {
    //                                  ^^ FIXME
    // There should not be a need for the size check as there should never be
    // any empty lists in the lookup_table.
    return empty_wordlist;
  } else {
    M = (*i).key();
    return *i;
  }
}

void SC_Writable_Soundslikes::add(const char* w, const string &s) {
  lookup_table[s.c_str()].push_back(w);
}

void SC_Writable_Soundslikes::erase(const char* w, const string &s) {
  WordList& item = lookup_table.find(s.c_str())->second;
  WordList::iterator i = item.begin();
  WordList::iterator end = item.end();
  while(i != end && *i != w) ++i;
  if (i == end) return;
  item.erase(i);
  if (!item.size()) {
    lookup_table.erase(s.c_str());
  }
}

const SC_Writable_Soundslikes::WordList * SC_Writable_Soundslikes
::words_w_soundslike(const char *word) const {
  LookupTable::const_iterator i = lookup_table.find(word);
  if (i == lookup_table.end()) {
    return NULL;
  } else {
    return &(i->second);
  }
}

const SC_Writable_Soundslikes::WordList * SC_Writable_Soundslikes
::words_w_soundslike(const char *word, SC_Sound &M) const {
  LookupTable::const_iterator i = lookup_table.find(word);
  if (i == lookup_table.end()) {
    return NULL;
  } else {
    M = i->first.c_str();
    return &(i->second);
  }
}

//
//
//

struct ReadOnlyWL::Internal {
  string wl_file_name;
  string sl_file_name;
  SC_ReadOnly_WordList    words;
  SC_ReadOnly_Soundslikes soundslikes;
  Internal(ReadOnlyWL * w) : words(w), soundslikes(w) {}
};

const char * ReadOnlyWL::file_name() const {
  //FIXME:  This should return the base (file_name - ext) for the file name
  return internal->wl_file_name.c_str();
}

ReadOnlyWL::const_iterator ReadOnlyWL::begin() const {
  return make_virtual_forward_iteratorr<const char *>(internal->words.begin());
}

ReadOnlyWL::const_iterator ReadOnlyWL::end() const {
  return make_virtual_forward_iteratorr<const char *>(internal->words.end());
}

ReadOnlyWL::size_type ReadOnlyWL::size() const {
  return internal->words.size();
}

bool ReadOnlyWL::empty() const {
  return internal->words.empty();
}

ReadOnlyWL::ReadOnlyWL() {
  internal = new Internal(this);
}

ReadOnlyWL::~ReadOnlyWL() {
  delete internal;
}

void ReadOnlyWL::load(const string &b, const string &lang_dir) {
  internal->wl_file_name = find_data_file(b, ".wrd");
  internal->sl_file_name = find_data_file(b, ".sl");
  internal->words.load(internal->wl_file_name);
  internal->soundslikes.load(internal->sl_file_name, 
			     internal->words.get_word_block());
}

bool ReadOnlyWL::operator [] (const char *word) const {
  return internal->words[word];
}

bool ReadOnlyWL::operator [] (const string &word) const {
  return internal->words[word];
}

const ReadOnlyWL::WordList ReadOnlyWL::words_w_soundslike(const char *word) const {
  const SC_ReadOnly_Soundslikes::WordList 
    temp = internal->soundslikes.words_w_soundslike(word);
  return make_virtual_const_begin_end_sizer<WordList>(temp);
}

//
//
//

static time_t modification_date(const string &file) {
  struct stat file_stat;
  if (stat(file.c_str(), &file_stat) == 0)
    return file_stat.st_mtime;
  else 
    return 0;
}

//
//
//

struct WritableWL::Internal {
  string file_name;
  time_t file_date;
  WritableWL * real_wl_;
  SC_Writable_WordList    words;
  SC_Writable_Soundslikes soundslikes;
  static const SC_Writable_Soundslikes::WordList empty_wordlist;
  Internal(WritableWL *w) : real_wl_(w), words(w), soundslikes(w) {}
  void save_as(const string &);
  void merge(const string &, const string &lang_dir="");
};

const SC_Writable_Soundslikes::WordList WritableWL::Internal::empty_wordlist;

struct SC_WWL_deref_via_cstr {
  const char * operator() (SC_Writable_WordList::const_iterator i) const
    {return i->c_str();}
};

WritableWL::const_iterator WritableWL::begin() const {
  return make_virtual_forward_iteratorr<const char *>
    (alt_dereference_iterator<SC_Writable_WordList::const_iterator,
                              SC_WWL_deref_via_cstr,
                              const char *, const char *, const char **>
     (internal->words.begin())
    );
}

WritableWL::const_iterator WritableWL::end() const {
  return make_virtual_forward_iteratorr<const char *>
    (alt_dereference_iterator<SC_Writable_WordList::const_iterator,
                              SC_WWL_deref_via_cstr,
                              const char *, const char *, const char **>
     (internal->words.end())
    );
  //return make_virtual_forward_iteratorr<const char *>(internal->words.end());
}

WritableWL::size_type WritableWL::size() const {
  return internal->words.size();
}

bool WritableWL::empty() const {
  return internal->words.empty();
}

void WritableWL::Internal::merge(const string &b, const string &lang_dir) {
  string fn = find_per_file(b,".per");
  ifstream WORDS( fn.c_str() );
  if ( WORDS.rdstate() & ios::badbit )
    throw ae_cant_read(fn, real_wl_);

  int c;
  string word, sound;
  const char *word2;
  
  WORDS >> word;
  if (word != "personal_wl")
    throw ae_bad_format(fn, real_wl_);

  WORDS >> word;

  try {
    real_wl_->set_check_lang(word, lang_dir);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, real_wl_);
  }

  WORDS >> c; // not used at the moment
  
  for (;;) {
    WORDS >> word >> sound;
    if (!WORDS) break;
    word2 = words.add(word);
    soundslikes.add(word2, sound);
  }

}

void WritableWL::Internal::save_as(const string &b) {
  file_name = find_per_file(b, ".per");
  
  ofstream WORDS((file_name).c_str());

  if (WORDS.rdstate() & ios::badbit) 
    throw ae_cant_write(file_name, real_wl_);

  WORDS << "personal_wl" << ' ' << real_wl_->lang_name() <<  endl;
  WORDS << words.size() << endl;

  SC_Writable_Soundslikes::const_iterator i = soundslikes.begin();
  SC_Writable_Soundslikes::const_iterator e = soundslikes.end();

  SC_Writable_Soundslikes::WordList::const_iterator j;
  
  for (;i != e; ++i) {
    for (j = i->second.begin(); j != i->second.end(); ++j) {
      WORDS << *j << ' ' << i->first << endl;
    }
  }

  WORDS.close();
  file_date = modification_date(file_name);
  
}

WritableWL::WritableWL() {
  internal = new Internal(this);
}

WritableWL::~WritableWL() {
  delete internal;
}

void WritableWL::load(const string &b, const string &lang_dir) {
  internal->file_name = find_per_file(b, ".per");
  internal->file_date = modification_date(internal->file_name);
  internal->merge(internal->file_name, lang_dir);
}

void WritableWL::merge(const string &b) {
  internal->merge(b);
}

void WritableWL::update() {
  internal->file_date = modification_date(internal->file_name);
  internal->merge(internal->file_name);
}

void WritableWL::save_as(const string &b) {
  return internal->save_as(b);
}

void WritableWL::save() {
  time_t new_date = modification_date(internal->file_name);
  if (new_date > internal->file_date)
    update();
  internal->save_as(internal->file_name);
  internal->file_date = modification_date(internal->file_name);
}

const char * WritableWL::file_name() const {
  return internal->file_name.c_str();
}

void WritableWL::add(const string &w) {
  if (!operator[](w))
    internal->soundslikes.add(internal->words.add(w),
			      lang()->to_soundslike(w));
}

void WritableWL::add(const string &w, const string &s) {
  if (!operator[](w))
    internal->soundslikes.add(internal->words.add(w),s);
}

void WritableWL::clear() {
  internal->words.clear(); 
  internal->soundslikes.clear();
}

bool WritableWL::operator [] (const char *word) const {
  return internal->words[word];
}

bool WritableWL::operator [] (const string &word) const {
  return internal->words[word];
}

const WritableWL::WordList WritableWL::words_w_soundslike(const char *word) const {
  const SC_Writable_Soundslikes::WordList * 
    temp = internal->soundslikes.words_w_soundslike(word);
  if (temp) {
    return make_virtual_const_begin_end_sizer<WordList>(*temp);
  } else {
    return make_virtual_const_begin_end_sizer<WordList>(Internal::empty_wordlist);
  }
}

//
//
//

bool WritableReplList::RealReplacements::exists(const string &word) {
  iterator i = begin();
  iterator e = end();
  while (i != e) {
    if (*i == word) return true;
    ++i;
  }
  return false;
}

bool WritableReplList::RealReplacements::add(const string &word) {
  if (exists(word)) return false;
  info.push_back(word);
  return true;
}

bool WritableReplList::RealReplacements::erase(const string &word) {
  vector<string>::iterator i = info.begin() + 1;
  vector<string>::iterator e = info.end();
  while (i != e) {
    if (*i == word) {
      info.erase(i);
      return true;
    }
    ++i;
  }
  return false;
}

struct WritableReplList::Internal {
  typedef vector<RealReplacements> ReplList;
  typedef hash_map<string, ReplList, hash_string<string> > LookupTable;
  struct deref_via_cstr {
    const char * operator() (RealReplacements::const_iterator i) const
      {return i->c_str();}
  };
  typedef alt_deref_make_virtual_forward_iterator
     < RealReplacements::const_iterator, deref_via_cstr, 
       const char *, const char *, const char**>
     repl_vir_itr;
  struct deref_to_repl {
    Replacements operator() (ReplList::const_iterator i) const {
      return Replacements(i->misspelled_word().c_str(),
			  repl_vir_itr(i->begin()),
			  repl_vir_itr(i->end()),
			  i->size());
    }
  };
  typedef alt_deref_make_virtual_forward_iterator
            < ReplList::const_iterator, deref_to_repl, 
	      Replacements, Replacements, Replacements* > 
    repl_list_vir_itr;

  static const ReplList empty_wordlist0;
  static const WordList empty_wordlist;

  WritableReplList * real_wl_;
  string file_name;
  time_t file_date;
  LookupTable lookup_table;

  Internal(WritableReplList * w) : real_wl_(w) {}
  void save_as(const string &);
  void load(const string &, const string &lang_dir = "");
  void merge(const string &);
};

#define Inter WritableReplList::Internal
const Inter::ReplList            Inter::empty_wordlist0;
const WritableReplList::WordList Inter::empty_wordlist
  (repl_list_vir_itr(empty_wordlist0.begin()),
   repl_list_vir_itr(empty_wordlist0.end()),
   empty_wordlist0.size());
#undef Inter


WritableReplList::WritableReplList() {
  internal = new Internal(this);
}

WritableReplList::~WritableReplList() {
  delete internal;
}

void WritableReplList::add(const string &mis, const string &cor) {
  add(mis, cor, lang()->to_soundslike(mis));
}

void WritableReplList
::add(const string &mis, const string &cor, const string &s) {
  Internal::LookupTable::iterator i = internal->lookup_table.find(s);
  if (i == internal->lookup_table.end())
    i = internal->lookup_table.insert
          (Internal::LookupTable::value_type(s, Internal::ReplList())).first;
  
  Internal::ReplList::iterator    j = i->second.begin();
  Internal::ReplList::iterator    e = i->second.end();
  for (; j != e; ++j) {
    if ((*j).misspelled_word() == mis) {
      (*j).add(cor);
      return;
    }
  }
  i->second.push_back(RealReplacements(mis,cor));
}

const WritableReplList::WordList 
WritableReplList::words_w_soundslike(const char *word) const {
  Internal::LookupTable::const_iterator i = internal->lookup_table.find(word);
  if (i == internal->lookup_table.end()) {
    return Internal::empty_wordlist;
  } else {
    return WordList(Internal::repl_list_vir_itr(i->second.begin()),
		    Internal::repl_list_vir_itr(i->second.end()),
		    i->second.size());
  }
}

void WritableReplList::Internal::save_as(const string &b) {
  file_name = find_per_file(b,".rpl");
  ofstream WORDS((file_name).c_str());

  if (WORDS.rdstate() & ios::badbit) 
    throw ae_cant_write(file_name, real_wl_);

  WORDS << "personal_repl" << ' ' << real_wl_->lang_name() <<  endl;
  WORDS << lookup_table.size() << ' ' << lookup_table.bucket_count() << endl;
  
  LookupTable::iterator i = lookup_table.begin();
  LookupTable::iterator e = lookup_table.end();
  
  for (;i != e; ++i) {
    WORDS << i->first << ' ' << i->second.size() << endl;
    for (ReplList::iterator j = i->second.begin(); j != i->second.end(); ++j) {
      WORDS << "  " << (*j).misspelled_word() << ' ' << (*j).size() << ' ';
      for (RealReplacements::iterator k = j->begin(); k != j->end(); ++k) {
	WORDS << *k << ',';
      }
      WORDS << endl;
    }
  }

  WORDS.close();
  
  file_date = modification_date(file_name);
}

void WritableReplList::save_as(const string &b) {
  internal->save_as(b);
}

void WritableReplList::save() {
  time_t new_date = modification_date(internal->file_name);
  if (new_date > internal->file_date)
    update();
  internal->save_as(internal->file_name);
  internal->file_date = modification_date(internal->file_name);
}

void WritableReplList::Internal::load(const string &b, const string &lang_dir) {
  file_name = find_per_file(b, ".rpl");
  file_date = modification_date(file_name);
  ifstream WORDS( (file_name).c_str() );
  if ( WORDS.rdstate() & ios::badbit )
    throw ae_cant_read(file_name, real_wl_);

  int c;
  string word, mis, sound, repl;
  int num_words, num_repls;

  WORDS >> word;
  if (word != "personal_repl")
    throw ae_bad_format(file_name, real_wl_);

  WORDS >> word;

  try {
    real_wl_->set_check_lang(word, lang_dir);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, real_wl_);
  }

  int num_soundslikes;
  WORDS >> num_soundslikes;
  WORDS >> c;  // not used at the moment

  int h,i,j;
  for (h=0; h != num_soundslikes; ++h) {
    WORDS >> sound >> num_words;
    ReplList &wl = lookup_table.insert(
      LookupTable::value_type(sound,ReplList())).first->second;
    wl.reserve(num_words);
    for (i = 0; i != num_words; ++i) {
      WORDS >> mis >> num_repls;
      WORDS.ignore(); // ignore space
      wl.push_back(RealReplacements(mis,num_repls));
      RealReplacements &rp = *(wl.end()-1);
      for (j = 0; j != num_repls; ++j) {
	getline(WORDS, repl, ',');
	rp.add_nocheck(repl);
      }
    }
  }
}

void WritableReplList::Internal::merge(const string &b) {

  string fn = find_per_file(b, ".rpl");

  ifstream WORDS( (fn).c_str() );
  if ( WORDS.rdstate() & ios::badbit ) 
    throw ae_cant_read(fn, real_wl_);

  int c;
  string word, mis, sound, repl;
  int num_words, num_repls;

  WORDS >> word;
  if (word != "personal_repl")
    throw ae_bad_format(fn, real_wl_);

  WORDS >> word;

  try {
    real_wl_->check_lang(word);
  } catch (ae_language_problem& e) {
    e.rethrow_w_file(file_name, real_wl_);
  }

  int num_soundslikes;
  WORDS >> num_soundslikes;
  WORDS >> c;  // not used at the moment

  int h,i,j;
  ReplList::iterator k;
  bool found;
  RealReplacements * rp;
  for (h=0; h != num_soundslikes; ++h) {
    WORDS >> sound >> num_words;
    ReplList &wl = lookup_table.insert(
      LookupTable::value_type(sound,ReplList())).first->second;
    for (i = 0; i != num_words; ++i) {
      WORDS >> mis >> num_repls;
      WORDS.ignore(); // ignore space
      found = false;
      for (k = wl.begin(); k != wl.end(); ++k) {
	if (mis == (*k).misspelled_word()) {
	  found = true;
	  break;
	}
      }
      if (found) {
	rp = &(*k);
      } else {
	wl.push_back(RealReplacements(mis,num_repls));
        rp = &(*(wl.end()-1));
      }
      for (j = 0; j != num_repls; ++j) {
	getline(WORDS,repl,';');
	rp->add(repl);
      }
    }
  }
}

void WritableReplList::load(const string &b, const string &lang_dir) {
  internal->load(b, lang_dir);
}

void WritableReplList::merge(const string &b) {
  internal->merge(b);
}

void WritableReplList::update() {
  internal->file_date = modification_date(internal->file_name);
  internal->merge(internal->file_name);
}

const char * WritableReplList::file_name() {
  return internal->file_name.c_str();
}


